import json
from pathlib import Path

part_types = {
    'Balun': {'subcategory': {}},
    'Battery': {'subcategory': {}},
    'Battery Holder': {'subcategory': {}},
    'Bolt': {'subcategory': {}},
    'Bridge Rectifier': {'subcategory': {}},
    'Common Mode Choke': {'subcategory': {}},
    'Capacitor': {
        'subcategory': {
            'Aluminium Electrolytic Capacitor': {'required_parameters': []},
            'MLCC': {'required_parameters': []},
        }
    },
    'Connector': {
        'required_parameters': [],
        'subcategory': {
            'Connector Bus': {'required_parameters': []},
            'Connector Accessory': {'required_parameters': []},
            'Connector IDC': {'required_parameters': []},
            'Connector Terminal Block': {'required_parameters': []},
            'Connector microSD Card': {'required_parameters': []},
            'Connector Pins': {'required_parameters': []}
        },
    },
    'Crystal': {
        'required_fields': [],
        'required_parameters': ['Frequency'],
        'subcategory': {}
    },
    'Crystal Oscillator': {'subcategory': {}},
    'Display': {
        'subcategory': {
            'LCD Display': {'required_parameters': []}
        }
    },
    'Diode': {
        'required_parameters': [],
        'subcategory': {
            'Schottky Diode': {'required_parameters': []},
            'Small Signal Diode': {'required_parameters': []},
            'Zener Diode': {'required_parameters': []},
            'LED': {'required_parameters': []},
            'TVS': {'required_parameters': []}
        }
    },
    'Enclosure': {'subcategory': {}},
    'Enclosure Accessory': {'subcategory': {}},
    'ESD Suppressor': {'subcategory': {}},
    'Fuse': {
        'subcategory': {
            'PTC Fuse': {'required_parameters': []}
        }
    },
    'IC': {
        'required_fields': ["description"],
        'required_parameters': [],
        'subcategory': {
            'IC Voltage Reference': {'required_parameters': []},
            'IC LDO': {'required_parameters': []},
            'IC Voltage Regulator': {'required_parameters': []},
            'IC Voltage Regulator Switching': {'required_parameters': []},
            'IC MCU': {'required_parameters': []},
            'IC Comparator': {'required_parameters': []},
            'IC Opamp': {'required_parameters': []},
            'IC Level translator': {'required_parameters': []},
            'IC Current Sense': {'required_parameters': []},
            'IC Load Switch': {'required_parameters': []},
            'IC RF Amplifier': {'required_parameters': []},
            'IC RF Synthesizer': {'required_parameters': []},
            'IC DAC': {'required_parameters': []},
            'IC ADC': {'required_parameters': []},
            'IC Sensor': {'required_parameters': []},
        }
    },
    'Inductor': {
        'subcategory': {
        }
    },
    'Lightpipe': {
        'required_fields': ["description"],
        'subcategory': {
        }
    },
    'Module': {
        'required_fields': ["description"],
        'subcategory': {}
    },
    'Resistor': {
        'subcategory': {
            'Resistor Thick Film': {'required_parameters': []},
            'Resistor Thin Film': {'required_parameters': []},
            'Resistor Array': {'required_parameters': []}
        }
    },
    'Relay': {'subcategory': {}},
    'Transistor': {
        'required_parameters': [],
        'subcategory': {
            'Transistor MOSFET P': {'required_parameters': []},
            'Transistor MOSFET N': {'required_parameters': []}
        }
    },
    'Surge arrester': {
        'subcategory': {}
    },
    'Switch': {
        'required_fields': ["description"],
        'subcategory': {}
    },
    'Varistor': {'subcategory': {}}
}


def load_part_types():
    part_types_dict = {}
    for item in part_types.keys():
        required_fields = part_types[item]['required_fields'] if 'required_fields' in part_types[item] else []
        required_parameters = part_types[item]['required_parameters'] if 'required_parameters' in part_types[item] else []
        part_types_dict[item] = {
            'required_fields': required_fields,
            'required_parameters': required_parameters
        }
        if 'subcategory' in part_types[item] and len(part_types[item]['subcategory']) > 0:
            for subitem_key in part_types[item]['subcategory']:
                subitem = part_types[item]['subcategory'][subitem_key]
                part_types_dict[subitem_key] = {
                    'required_fields': required_fields + subitem['required_fields'] if 'required_fields' in subitem else [],
                    'required_parameters': required_parameters +  subitem['required_parameters'] if 'required_parameters' in subitem else []
                }
    return set(part_types_dict.keys()), part_types_dict


def load_manufacturers(path: Path):
    with open(path) as f:
        manufacturers = json.load(f)
        return [x['name'] for x in manufacturers] + [x['full_name'] for x in manufacturers if
                                                     x['full_name'] is not None and len(x['full_name']) > 0]


def load_packaging_types():
    return ['Bag', 'Bulk', 'Cut Tape', 'Embossed Tape / Reel', 'Paper Tape / Reel', 'Foil',
            'shrink wrap', 'Tube', 'Tray', '', 'Tape & Reel', '13” Reel', '7" reel']


def load_msl_classification():
    return ['MSL-1 UNLIM',
            'MSL-2 1-YEAR',
            'MSL-2A 4-WEEKS',
            'MSL-3 168-HOURS',
            'MSL-4 72-HOURS',
            'MSL-5 48-HOURS',
            'MSL-5A 24-HOURS',
            'MSL-6 TOL']


def load_files(directory):
    files = Path(directory).rglob('*.json')
    return [x for x in files if not str(x).endswith('_generated.json')]



import argparse
import json
from pathlib import Path
from .common import load_files


def main():
    parser = argparse.ArgumentParser(description='Part Database Tools')
    parser.add_argument('-f', '--file', type=Path, help='Path to the input file')
    parser.add_argument('-d', '--directory', type=Path, help='Path to the input directory')
    args = parser.parse_args()

    if args.file:
        if args.file.is_file():
            files = [args.file]
        else:
            print('Input file must be a file')
            return -1
    elif args.directory:
        files = load_files(args.directory.resolve())
    else:
        print('Missing parameter. Input file or directory must be specified')
        return -1

    for f in files:
        cleanup_and_sort_json_files(f)
    return 0


def cleanup_json_data(data):
    if isinstance(data, dict):
        return {k: cleanup_json_data(v) for k, v in data.items() if v != ""}
    elif isinstance(data, list):
        return [cleanup_json_data(e) for e in data]
    else:
        return data


def cleanup_and_sort_json_files(f):
    try:
        with open(str(f)) as jsonfile:
            content = json.load(jsonfile)
            cleanup_json = cleanup_json_data(content)
            sorted_json = sorted(cleanup_json, key=lambda x: x['manufacturer'] + x['partNumber'])
            fixed_content = json.dumps(sorted_json, indent='\t')
            jsonfile.close()
            if content != fixed_content:
                with open(str(f), 'w') as jsonfile:
                    jsonfile.write(fixed_content)
    except ValueError as e:
        print(f"Invalid JSON file {f}: {e}")



from ..common import load_packaging_types

order_number_allowed_fields = ['12NC', 'Alias', 'Status', 'Code', 'Type', 'Qty', 'PackagingData']

def validate_order_number_fields(part):
    for order_number in part['orderNumbers']:
        order = part['orderNumbers'][order_number]
        for field in order:
            if field not in order_number_allowed_fields:
                return False

        if not validate_packaging_type(order):
            return False
    return True


def validate_packaging_type(order_number):
    if 'Type' in order_number:
        if order_number['Type'] not in load_packaging_types():
            return False
    return True


import argparse
import json

from pathlib import Path
from .common import load_manufacturers
from .common import load_files, load_part_types, load_msl_classification
from .validators.orderNumberValidator import validate_order_number_fields

msl_classification = load_msl_classification()

def main():
    parser = argparse.ArgumentParser(description='Validate the parts database')
    parser.add_argument('-m', '--manufacturers', type=Path, required=True, help="Manufacturers definition file")
    parser.add_argument('-i', '--input', type=Path, help="Input file")
    parser.add_argument('-d', '--directory', type=Path, help="Input directory")
    args = parser.parse_args()

    manufacturers = load_manufacturers(args.manufacturers)
    part_types, parameters = load_part_types()
    if args.input:
        files = [args.input]
    elif args.directory:
        files = load_files(args.directory)

    json_files = []
    for f in files:
        with open(str(f)) as jsonfile:
            json_files.append({'file': f, 'data': json.load(jsonfile)})

    for f in json_files:
        parts = f['data']
        for part in parts:
            if not validate_generic_required_fields(part):
                print("missing required field")

            if not validate_manufacturer(part, manufacturers):
                print("invalid manufacturer", part["manufacturer"])

            if not validate_part_type(part, part_types):
                print("invalid part type", part["partType"])

            if not validate_field_storage_condition(part):
                print("invalid field storage condition", part["storageConditions"])

            if not test_validate_files_field(part):
                print("invalid file field", part["files"])

            if not validate_order_number_fields(part):
                print("invalid order number field", part["order_number"])


def validate_generic_required_fields(part):
    required_fields = ["manufacturer", "partNumber", "partType"]
    for required_field in required_fields:
        if required_field not in part:
            return False
    return True


def validate_manufacturer(part, manufacturer):
    if part['manufacturer'] in  manufacturer:
        return True
    return False


def validate_part_type(part, part_types):
    if part['partType'] in part_types:
        return True
    return False

def validate_field_storage_condition(part):
    if 'storageConditions' in part:
        if 'MSLevel' not in part['storageConditions']:
            print(f"Missing MSLevel field in part {part['manufacturer']} {part['partNumber']}")
            return False
        if part['storageConditions']['MSLevel'] not in msl_classification:
            return False
    return True

def test_validate_files_field(part):
    if 'files' in part:
        if not isinstance(part['files'], dict):
            return False
        for attachment in part['files']:
            if 'url' not in part['files'][attachment]:
                return False
            # check if only allowed keys are present in file dict
            for key in part['files'][attachment]:
                if key not in ["filename", "url", "description", "versions"]:
                    print("Key not allowed", key)
                    return False
    return True



import argparse
import cv2 as cv
import numpy as np

from pathlib import Path
from .common import load_manufacturers

min_size_x = 640*2
min_size_y = 480*2
small_object_size_x=640*2
small_object_size_y=480*2

class OutputImageParameters:
    def __init__(self, width: int, height: int, border: int):
        assert width > 0 and height > 0
        assert border >= 0
        assert width - 2*border > 0
        assert height - 2*border > 0
        self.width = width
        self.height = height
        self.border = border
        self.ratio = width/height
        self.object_width = width - 2*border
        self.object_height = height - 2*border

class BoundingBox:
    def __init__(self, xmin, xmax, ymin, ymax, image_width, image_height):
        assert xmin < xmax
        assert ymin < ymax
        assert xmax <= image_width, f"{xmax} > {image_width}"
        assert ymax <= image_height, f"{ymax} > {image_height}"
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        self.size_x = xmax - xmin
        self.size_y = ymax - ymin
        self.ratio = self.size_x/self.size_y
        self.image_width = image_width
        self.image_height = image_height

    def center(self):
        return self.xmin + (self.xmax - self.xmin) / 2, self.ymin + (self.ymax - self.ymin) / 2

    def __str__(self):
        return f"Bounding box {self.xmin}-{self.xmax}-{self.ymin}-{self.ymax}"


def main():
    parser = argparse.ArgumentParser(description="Picture tool")
    parser.add_argument('-b', '--background_image', type=Path, help="Background image")
    parser.add_argument('-i', '--input_dir', type=Path, required=True, help="Input image directory")
    parser.add_argument('-o', '--output_dir', type=Path, required=True, help="Output image directory")
    parser.add_argument('-d', '--debug', action='store_true', help="Enable debug mode")
    parser.add_argument('-m', '--manufacturers', type=Path, help="Manufacturers definition file")
    parser.add_argument('--output_image_size_x', type=int, default=640, help="Output image size")
    parser.add_argument('--output_image_size_y', type=int, default=480, help="Output image size")
    args = parser.parse_args()

    out_img_params = OutputImageParameters(args.output_image_size_x, args.output_image_size_y, 20)
    if args.manufacturers is not None:
        manufacturers = load_manufacturers(args.manufacturers.resolve())
    else:
        manufacturers = None

    image_collections = load_images(args.input_dir, args.background_image)
    for collection in image_collections:
        try:
            print("Processing", collection)
            processed_images = process_images(
                image_collections[collection]["background"],
                image_collections[collection]["images"],
                out_img_params,
                debug=args.debug
            )
            for processed_image_name in processed_images:
                save_image(processed_images[processed_image_name], args.output_dir, processed_image_name, manufacturers)
        except Exception as e:
            print(f"Exception while processing {collection}: ", e)


def process_images(background, images, out_image_params, debug=False):
    back_sub = cv.createBackgroundSubtractorMOG2(history=10, varThreshold=200, detectShadows=False)
    for i in range(4):
        back_sub.apply(background)

    fg_masks = []
    for image_name in images:
        fg_masks.append(back_sub.apply(images[image_name]))
        if debug:
            cv.imshow('FG Mask', cv.resize(fg_masks[-1], (640, 480), interpolation=cv.INTER_AREA))
            cv.waitKey(300)

    fg_mask = cv.bitwise_or(filterout_small_elements(fg_masks[-2]), filterout_small_elements(fg_masks[-1]))
    if debug:
        cv.imshow('FG Mask sum', cv.resize(fg_mask, (640, 480), interpolation=cv.INTER_AREA))
        cv.waitKey(300)
    bounding_box = get_bounding_box(fg_mask)

    processed_images = {}
    if is_small_object(bounding_box):
        dims = inside_min_image(bounding_box)
        x1, x2, y1, y2 = dims
        for image_name in images:
            final_bounding_box = BoundingBox(x1 - out_image_params.border,
                                             x2 + out_image_params.border,
                                             y1 - out_image_params.border,
                                             y2 + out_image_params.border,
                                             images[image_name].shape[1],
                                             images[image_name].shape[0])
            output_image = crop_image(images[image_name], final_bounding_box)
            resized_img = cv.resize(
                output_image,
                (out_image_params.width, out_image_params.height),
                interpolation=cv.INTER_AREA)
            processed_images[image_name] = resized_img
    else:
        for image_name in images:
            final_bounding_box = bounding_box_to_proportion(bounding_box, out_image_params)
            cropped_image=crop_image(images[image_name], final_bounding_box)
            resized_img = cv.resize(
                cropped_image,
                (out_image_params.width, out_image_params.height),
                interpolation=cv.INTER_AREA)
            processed_images[image_name] = resized_img
    return processed_images


def load_images(input_dir: Path, background_image: Path):
    image_collection = {}
    for file in input_dir.glob('**/*'):
        if file.is_file():
            filename = file.stem
            basename_end = filename.rfind("__")
            if basename_end != -1:
                collection = filename[:basename_end]
                if collection not in image_collection:
                    image_collection[collection] = {"background": None, "images": {}}
                if filename.endswith("background"):
                    image_collection[collection]["background"] = cv.imread(file.resolve())
                else:
                    image_collection[collection]["images"][filename] = cv.imread(file.resolve())

    if background_image is not None and background_image.is_file():
        background_img = cv.imread(background_image.resolve())
        for collection in image_collection:
            if image_collection[collection]["background"] is None:
                image_collection[collection]["background"] = background_img
    return image_collection


def save_image(image, output_dir, filename, manufacturers):
    manufacturer_name = filename.split("__")[0]
    if manufacturer_name in manufacturers or manufacturers is None:
        write_path = output_dir.joinpath(manufacturer_name).joinpath(filename).with_suffix(".jpg")
        write_path.parent.mkdir(parents=True, exist_ok=True)
        status = cv.imwrite(write_path.resolve(), image)
        if not status:
            print("Error writing image:", write_path.resolve())


def is_small_object(bounding_box):
    if bounding_box.size_x <= small_object_size_x and bounding_box.size_y <= small_object_size_y:
        return True
    else:
        return False


def filterout_small_elements(fg_mask):
    nb_blobs, im_with_separated_blobs, stats, _ = cv.connectedComponentsWithStats(fg_mask)
    sizes = stats[:, cv.CC_STAT_AREA]
    min_size = 1000

    im_result = np.zeros_like(im_with_separated_blobs, dtype=np.uint8)
    for index_blob in range(1, nb_blobs):
        if sizes[index_blob] >= min_size:
            im_result[im_with_separated_blobs == index_blob] = 255
    return im_result


def bounding_box_to_proportion(bounding_box, image_parameters: OutputImageParameters):
    if bounding_box.ratio > image_parameters.ratio:
        # add borders and adjust heigh
        scaling_ratio = bounding_box.size_x / image_parameters.object_width
        width_border = image_parameters.border * scaling_ratio
        if bounding_box.xmin - width_border < 0:
            width_border = bounding_box.xmin
            scaling_ratio = (bounding_box.size_x + 2*width_border) / image_parameters.width
        height = image_parameters.height * scaling_ratio
        if height > bounding_box.image_height:
            if bounding_box.image_width / bounding_box.image_height == image_parameters.ratio:
                return BoundingBox(0, bounding_box.image_width, 0, bounding_box.image_height, bounding_box.image_width, bounding_box.image_height)
            else:
                print(f"Original image ratio {bounding_box.image_width/bounding_box.image_height}, "
                      f"requested ratio {image_parameters.ratio}, scaling factor: {scaling_ratio}, scaled height: {height}")
                raise ValueError("Unable to fit bounding box with requested image ratio")
        center_y = bounding_box.center()[1]
        y_min = int(center_y - height / 2)
        return BoundingBox(int(bounding_box.xmin - width_border),
                           int(bounding_box.xmax + width_border),
                           y_min,
                           int(y_min + height),
                           bounding_box.image_width,
                           bounding_box.image_height)
    else:
        print("Adjusting width")
        # add borders and adjust width
        scaling_ratio = bounding_box.size_y / image_parameters.object_height
        height_border = image_parameters.border * scaling_ratio
        if bounding_box.ymin - height_border < 0:
            height_border = bounding_box.ymin
            scaling_ratio = (bounding_box.size_y + 2*height_border) / image_parameters.height
        width_border = image_parameters.border * scaling_ratio

    return BoundingBox(int(bounding_box.xmin - width_border),
                       int(bounding_box.xmax + width_border),
                       int(bounding_box.ymin * scaling_ratio - height_border),
                       int(bounding_box.ymax * scaling_ratio + height_border),
                       bounding_box.image_width,
                       bounding_box.image_height)


def get_bounding_box(elements):
    pts = np.argwhere(elements>0)
    y1, x1 = pts.min(axis=0)
    y2, x2 = pts.max(axis=0)
    return BoundingBox(x1, x2, y1, y2, elements.shape[1], elements.shape[0])


def crop_image(image, bounding_box):
    return image[
           bounding_box.ymin:bounding_box.ymax,
           bounding_box.xmin:bounding_box.xmax]


def inside_min_image(bounding_box):
    if bounding_box.size_x <= min_size_x and bounding_box.size_y <= min_size_y:
        middle_x = int(bounding_box.xmin + bounding_box.size_x / 2)
        middle_y = int(bounding_box.ymin + bounding_box.size_y / 2)
        return middle_x - min_size_x, middle_x + min_size_x, middle_y - min_size_y, middle_y + min_size_y
    return None



import argparse
import hashlib
import shutil
import json

from pathlib import Path
import urllib.request
from urllib.parse import urlparse, unquote

from .common import load_files


class FileData:
    def __init__(self, filename, url, manufacturer, filetype, revision, date, md5sum, path):
        self.filename = Path(filename if filename is not None else self.filename_from_url(url))
        self.url = url
        self.manufacturer = manufacturer
        self.filetype = filetype
        self.revision = revision
        self.date = date
        self.md5sum = md5sum
        self.path = path

    def set_destination_directory(self, destination_directory):
        manufacturer = self.manufacturer.replace(" ", "_").lower()
        file_name = f"{manufacturer}__{self.filename.stem}__{self.md5sum}"
        self.path = destination_directory.joinpath(manufacturer, file_name).with_suffix(self.filename.suffix)

    def exists(self):
        return self.path.exists() and self.path.is_file()

    def validate(self):
        return self.md5sum == calculate_md5(self.path)

    def download(self):
        local_filename, headers = urllib.request.urlretrieve(self.url)
        md5sum = calculate_md5(local_filename)
        if md5sum == self.md5sum:
            self.path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(local_filename, self.path)

    def filename_from_url(self, url):
        return Path(unquote(urlparse(url).path).replace(' ', '_')).name



def main():
    parser = argparse.ArgumentParser(description="Part DB tools")
    parser.add_argument('-i', '--input', help="Path to the input file")
    parser.add_argument('-d', '--directory', help="Path to the input directory")
    parser.add_argument('-o', '--output', type=Path, required=True, help="Path to the output directory")
    parser.add_argument('-v', '--validate_only', action='store_true', help="Validate files")
    args = parser.parse_args()

    if args.input:
        input_files = [args.input]
    elif args.directory:
        input_files = load_files(args.directory)
    else:
        print("No input or directory specified")
        exit(-1)

    part_files = load_part_files(input_files)

    for f in part_files:
        f.set_destination_directory(args.output)
        if f.exists():
            print("file already exists, validating")
            if not f.validate():
                print("file validation failed")
        else:
            print(f"File missing: {f.path}")
            if not args.validate_only:
                print("Downloading")
                f.download()


def load_part_files(parts_files):
    files = []
    for part_file in parts_files:
        with open(part_file) as jsonfile:
            parts = json.load(jsonfile)
            for part in parts:
                for file_type in part["files"]:
                    file = part["files"][file_type]
                    if 'skip' in file["versions"] and file["versions"]["skip"] == True:
                        print("Skipping file, versions not supported")
                    else:
                        for version in file["versions"]:
                            file_data = FileData(
                                filename=file["filename"] if "filename" in file else None,
                                url=file["url"],
                                manufacturer=part["manufacturer"],
                                filetype=file_type,
                                revision=version,
                                date=file["versions"][version]["date"],
                                md5sum=file["versions"][version]["md5sum"],
                                path=None
                            )
                            files.append(file_data)
    return files


def calculate_md5(file):
    with open(file, "rb") as f:
        hash_md5 = hashlib.md5(f.read()).hexdigest()
        return hash_md5

def copy_and_rename_file(input_file: Path, output_dir: Path):
    manufacturer = input_file.stem.split("__")[0]
    file_name = input_file.stem.split("__")[1]
    md5_sum = calculate_md5(input_file)

    new_filename = f"{manufacturer}__{file_name}__{md5_sum}.{input_file.suffix}"
    new_file_path = output_dir.joinpath(new_filename)

    shutil.copyfile(input_file.resolve(), new_file_path.resolve())